pub mod introspection;
pub mod login;
pub mod oauth;
pub mod session;

pub mod auth {

    use std::{ops::Deref, sync::Arc};

    use anyhow::Context;
    use async_trait::async_trait;
    use axum::http::{header::SET_COOKIE, HeaderMap};
    use oauth2::url::Url;

    use crate::{
        introspection::IntrospectionService,
        login::{config::AuthEngine, AuthClap},
        oauth::{OAuth, ZitadelConfig},
        session::{SessionService, User},
    };

    #[async_trait]
    pub trait Auth {
        async fn login(&self) -> anyhow::Result<Url>;
        async fn login_token(&self, user: &str, password: &str) -> anyhow::Result<String>;
        async fn login_authorized(
            &self,
            code: &str,
            state: &str,
        ) -> anyhow::Result<(HeaderMap, Url)>;
        async fn get_user_from_session(&self, cookie: &str) -> anyhow::Result<User>;
    }

    #[derive(Clone)]
    pub struct AuthService(Arc<dyn Auth + Send + Sync + 'static>);

    impl AuthService {
        pub async fn new(config: &AuthClap, session: SessionService) -> anyhow::Result<Self> {
            match config.engine {
                AuthEngine::Noop => Ok(Self::new_noop()),
                AuthEngine::Zitadel => {
                    let oauth: OAuth = ZitadelConfig::try_from(config.zitadel.clone())?.into();
                    let introspection: IntrospectionService =
                        IntrospectionService::new_zitadel(config).await?;

                    Ok(Self::new_zitadel(oauth, introspection, session))
                }
            }
        }

        pub fn new_zitadel(
            oauth: OAuth,
            introspection: IntrospectionService,
            session: SessionService,
        ) -> Self {
            Self(Arc::new(ZitadelAuthService {
                oauth,
                introspection,
                session,
            }))
        }

        pub fn new_noop() -> Self {
            Self(Arc::new(NoopAuthService {}))
        }
    }

    impl Deref for AuthService {
        type Target = Arc<dyn Auth + Send + Sync + 'static>;

        fn deref(&self) -> &Self::Target {
            &self.0
        }
    }

    pub struct ZitadelAuthService {
        oauth: OAuth,
        introspection: IntrospectionService,
        session: SessionService,
    }
    pub static COOKIE_NAME: &str = "SESSION";

    #[async_trait]
    impl Auth for ZitadelAuthService {
        async fn login(&self) -> anyhow::Result<Url> {
            let authorize_url = self.oauth.authorize_url().await?;

            Ok(authorize_url)
        }
        async fn login_authorized(
            &self,
            code: &str,
            _state: &str,
        ) -> anyhow::Result<(HeaderMap, Url)> {
            let token = self.oauth.exchange(code).await?;
            let user_id = self.introspection.get_id_token(token.as_str()).await?;
            let cookie_value = self.session.insert_user("user", user_id.as_str()).await?;

            let cookie = format!("{}={}; SameSite=Lax; Path=/", COOKIE_NAME, cookie_value);

            let mut headers = HeaderMap::new();
            headers.insert(SET_COOKIE, cookie.parse().unwrap());

            Ok((
                headers,
                Url::parse("http://localhost:3000/dash/home")
                    .context("failed to parse login_authorized zitadel return url")?,
            ))
        }
        async fn login_token(&self, _user: &str, password: &str) -> anyhow::Result<String> {
            self.introspection.get_id_token(password).await
        }
        async fn get_user_from_session(&self, cookie: &str) -> anyhow::Result<User> {
            match self.session.get_user(cookie).await? {
                Some(u) => Ok(User { id: u }),
                None => Err(anyhow::anyhow!("failed to find user")),
            }
        }
    }

    pub struct NoopAuthService {}

    #[async_trait]
    impl Auth for NoopAuthService {
        async fn login(&self) -> anyhow::Result<Url> {
            todo!()
        }
        async fn login_authorized(
            &self,
            _code: &str,
            _state: &str,
        ) -> anyhow::Result<(HeaderMap, Url)> {
            todo!()
        }

        async fn login_token(&self, _user: &str, _password: &str) -> anyhow::Result<String> {
            todo!()
        }

        async fn get_user_from_session(&self, _cookie: &str) -> anyhow::Result<User> {
            todo!()
        }
    }
}

#[cfg(test)]
mod test {
    use clap::Parser;
    use pretty_assertions::assert_eq;

    use crate::{
        login::{
            config::{AuthEngine, ZitadelClap},
            AuthClap,
        },
        session::{PostgresqlSessionClap, SessionBackend, SessionClap},
    };

    #[derive(Parser)]
    #[command(author, version, about, long_about = None)]
    pub struct Cli {
        #[command(subcommand)]
        command: Commands,
    }

    #[derive(clap::Subcommand, Clone, Debug, Eq, PartialEq)]
    pub enum Commands {
        One {
            #[clap(flatten)]
            options: AuthClap,
        },
    }

    #[test]
    fn test_command_parse_as_default_noop() {
        let cli: Cli = Cli::parse_from(["base", "one"]);

        assert_eq!(
            cli.command,
            Commands::One {
                options: AuthClap {
                    engine: AuthEngine::Noop,
                    zitadel: ZitadelClap {
                        auth_url: None,
                        client_id: None,
                        client_secret: None,
                        redirect_url: None,
                        token_url: None,
                        authority_url: None,
                    },
                    session_backend: SessionBackend::InMemory,
                    session: SessionClap {
                        postgresql: PostgresqlSessionClap { conn: None }
                    }
                }
            }
        );
    }

    #[test]
    fn test_command_parse_as_noop() {
        let cli: Cli = Cli::parse_from(["base", "one", "--auth-engine", "noop"]);

        assert_eq!(
            cli.command,
            Commands::One {
                options: AuthClap {
                    engine: AuthEngine::Noop,
                    zitadel: ZitadelClap {
                        auth_url: None,
                        client_id: None,
                        client_secret: None,
                        redirect_url: None,
                        token_url: None,
                        authority_url: None,
                    },
                    session_backend: SessionBackend::InMemory,
                    session: SessionClap {
                        postgresql: PostgresqlSessionClap { conn: None }
                    }
                }
            }
        );
    }

    #[test]
    fn test_command_parse_as_zitadel() {
        let cli: Cli = Cli::parse_from([
            "base",
            "one",
            "--auth-engine=zitadel",
            "--zitadel-client-id=something",
            "--zitadel-client-secret=something",
            "--zitadel-auth-url=https://something",
            "--zitadel-redirect-url=https://something",
            "--zitadel-token-url=https://something",
            "--zitadel-authority-url=https://something",
        ]);

        assert_eq!(
            cli.command,
            Commands::One {
                options: AuthClap {
                    engine: AuthEngine::Zitadel,
                    zitadel: ZitadelClap {
                        auth_url: Some("https://something".into()),
                        client_id: Some("something".into()),
                        client_secret: Some("something".into()),
                        redirect_url: Some("https://something".into()),
                        token_url: Some("https://something".into()),
                        authority_url: Some("https://something".into()),
                    },
                    session_backend: SessionBackend::InMemory,
                    session: SessionClap {
                        postgresql: PostgresqlSessionClap { conn: None }
                    }
                },
            }
        );
    }
}
